from google.cloud.asset_v1 import AssetServiceClient
from google.cloud.bigquery import Client as BigQueryClient
from google.oauth2.credentials import Credentials
from rich.console import Console

from bq_meta.config import Config


class Client:
    def __init__(self, console: Console, config: Config):
        self._bq_client = None
        self._asset_client = None
        self._connection_client = None
        self._location_client = None
        self.console = console
        self.config = config

    @property
    def bq_client(self) -> BigQueryClient:
        if not self._bq_client:
            self._bq_client = BigQueryClient(
                project="",  # leaving project as None would result in error
                credentials=Credentials.from_authorized_user_info(self.config.credentials),
            )
        return self._bq_client

    @property
    def asset_client(self) -> AssetServiceClient:
        if not self._asset_client:
            self._asset_client = AssetServiceClient(
                credentials=Credentials.from_authorized_user_info(self.config.credentials),
            )
        return self._asset_client



import os
from typing import Optional

from google.cloud import bigquery
from jinja2 import Template
from loguru import logger

from bq_meta import const


class SnippetService:
    def __init__(self) -> None:
        self.projects_path = const.BQ_META_PROJECTS

    def list_snippets(self):
        logger.trace("Method call")
        return next(os.walk(const.BQ_META_SNIPPETS), (None, None, []))[2]

    def get_snippet(self, snippet: str, table: bigquery.Table) -> Optional[str]:
        logger.trace("Method call")
        rendered = None
        if snippet:
            with open(f"{const.BQ_META_SNIPPETS}/{snippet}") as f:
                rendered = Template(f.read()).render(
                    project=table.project,
                    dataset=table.dataset_id,
                    table=table.table_id,
                )
        logger.debug(f"Rendered snippet: {const.BQ_META_SNIPPETS}/{snippet}")
        return rendered



import os
from typing import Optional

import click
from rich.console import Console
from rich.panel import Panel
from rich.text import Text
from google.cloud.bigquery.table import TableReference

from bq_meta import const, output
from bq_meta.client import Client
from bq_meta.config import Config
from bq_meta.initialize import initialize
from bq_meta.service.history_service import HistoryService
from bq_meta.service.iam_service import IamService
from bq_meta.service.project_service import ProjectService
from bq_meta.service.table_service import TableService
from bq_meta.service.snippet_service import SnippetService
from bq_meta.service.version_service import VersionService
from bq_meta.util import table_utils
from bq_meta.window import Window
from loguru import logger
import os


@click.command()
@click.argument("full_table_id", required=False)
@click.option("--raw", help="View raw response from the BigQuery for specific 'FULL_TABLE_ID'", is_flag=True)
@click.option("--init", help="Initialize 'bq-meta' configuration", is_flag=True)
@click.option("--info", help="Print info of currently used account", is_flag=True)
@click.option("--fetch-projects", help="Fetch available google projects", is_flag=True)
@click.option("--purge-history", help="Remove non existing tables from history", is_flag=True)
@click.option("--debug", help="Log debug messages into BQ_META_HOME/debug.log", is_flag=True)
@click.option("--trace", help="Log tace messages into BQ_META_HOME/trace.log", is_flag=True)
@click.version_option()
def cli(
    full_table_id: Optional[str],
    raw: bool,
    init: bool,
    info: bool,
    fetch_projects: bool,
    purge_history: bool,
    debug: bool,
    trace: bool,
):
    """BiqQuery metadata"""

    ctx = click.get_current_context()
    console = Console(theme=const.theme, soft_wrap=True, force_interactive=True)
    config = Config()
    client = Client(console, config)
    project_service = ProjectService(console, config, client)
    table_service = TableService(console, config, client, project_service)
    snippet_service = SnippetService()
    history_service = HistoryService(console, config, table_service)
    iam_service = IamService(console, config, client)
    window = Window(console, config, history_service, table_service, snippet_service, iam_service)
    table = None

    logger.remove(0)  # disable stdout handler
    logger.level("DEBUG", color="<fg #008720>")
    logger.level("TRACE", color="<fg #6CEC16>")
    if debug:
        logger.add(const.BQ_META_DEBUG, colorize=True, level="DEBUG", rotation="1 MB", format=const.logger_format)
    if trace:
        logger.add(const.BQ_META_TRACE, colorize=True, level="TRACE", rotation="1 MB", format=const.logger_format)

    if os.path.exists(const.BQ_META_HOME):
        version_service = VersionService()
        version_service.update_config(config)

    logger.debug(f"Loaded config: {config.conf}")

    if init:
        initialize(config, console, project_service)
        ctx.exit()
    elif not os.path.exists(const.BQ_META_HOME):
        panel = Panel(
            title="Not initialized, run",
            renderable=Text("bq-meta --init"),
            expand=False,
            padding=(1, 3),
            border_style=const.request_style,
        )
        console.print(panel)
        ctx.exit()
    elif info:
        console.print(output.get_config_info(config))
        ctx.exit()
    elif fetch_projects:
        project_service.fetch_projects()
        ctx.exit()
    elif purge_history:
        history_service.purge_tables()
        ctx.exit()
    elif full_table_id:
        table_ref = TableReference.from_string(full_table_id.replace(":", "."))
        project_id = table_ref.project
        dataset_id = table_ref.dataset_id
        table_id = table_ref.table_id
        table = table_service.get_table(project_id, dataset_id, table_id)
        if raw:
            console.print_json(table_utils.get_properties(table))
            ctx.exit()

    window.live_window(table)



import webbrowser
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import List, Optional

import click
import pyperclip
import readchar
from google.cloud import bigquery
from loguru import logger
from readchar import key
from rich.console import Console, RenderableType
from rich.layout import Layout
from rich.live import Live

from bq_meta import output
from bq_meta.config import Config
from bq_meta.service.history_service import HistoryService
from bq_meta.service.iam_service import IamService
from bq_meta.service.snippet_service import SnippetService
from bq_meta.service.table_service import TableService
from bq_meta.util import table_utils
from bq_meta.util.rich_utils import flash_content, flash_panel


class View(Enum):
    empty = 1
    table = 2
    schema = 3
    snippets = 4
    metadata = 5
    iam = 6
    error = 7


class Metadata(Enum):
    table_id = 1
    link = 2
    schema = 3


@dataclass
class Hint:
    key: str
    name: str


hint_open = Hint("o", "Open new table")
hint_table = Hint("t", "Show table")
hint_schema = Hint("s", "Show schema")
hint_snippets = Hint("p", "Show snippets")
hint_metadata = Hint("m", "Show metadata")
hint_iam = Hint("i", "Show iam")
hint_history = Hint("h", "Show history")
hint_open_browser = Hint("b", "Open in browser")

hint_refresh = Hint("r", "Refresh")
hint_copy = Hint("c", "Copy")
hint_quit = Hint("q", "Quit")

all_hints = [
    hint_open,
    hint_open_browser,
    hint_table,
    hint_schema,
    hint_snippets,
    hint_metadata,
    hint_iam,
    hint_history,
]


class Window:
    def __init__(
        self,
        console: Console,
        config: Config,
        history_service: HistoryService,
        table_service: TableService,
        snippet_service: SnippetService,
        iam_service: IamService,
    ):
        self.console = console
        self.config = config
        self.history_service = history_service
        self.table_service = table_service
        self.snippet_service = snippet_service
        self.iam_service = iam_service
        self.table: Optional[bigquery.Table] = None
        self.project_id: Optional[str] = None
        self.dataset_id: Optional[str] = None
        self.layout: Layout = Layout()
        self.content: Optional[RenderableType] = None
        self.view: View = View.empty
        self.hints: List[str] = []
        self.bottom_hints: List[str] = [hint_quit]
        self.values: List[str] = []
        self.selected_value: Optional[str] = None
        self.snippet: Optional[str] = None

    def live_window(self, table: Optional[bigquery.Table]):
        self.now = datetime.utcnow()
        self.table = table
        with Live(self.layout, auto_refresh=False, screen=True, transient=True) as live:
            self._loop(live)

    def _update_content(self, live: Live):
        logger.trace("Updated content")
        match self.view:
            case View.empty:
                self.hints = [hint_open, hint_history]
            case View.table if self.table:
                self.hints = all_hints
                self.bottom_hints = [hint_refresh, hint_quit]
                self.content = output.get_table_output(self.table)
                self._update_table(self.table)
            case View.schema if self.table:
                live.stop()
                with self.console.pager():
                    self.console.print(output.get_schema_output(self.table))
                self.view = View.table
                live.start()
            case View.snippets if self.table:
                self.hints = all_hints
                self.bottom_hints = [hint_refresh, hint_copy, hint_quit]
                self.content = self.snippet_service.get_snippet(self.selected_value, self.table)
            case View.metadata if self.table:
                self.hints = all_hints
                self.bottom_hints = [hint_refresh, hint_copy, hint_quit]
                if self.selected_value == Metadata.table_id.name:
                    self.content = self.table.full_table_id
                elif self.selected_value == Metadata.link.name:
                    self.content = table_utils.get_table_link(self.table)
                elif self.selected_value == Metadata.schema.name:
                    self.content = table_utils.get_schema_json(self.table)
            case View.iam if self.table:
                self.hints = all_hints
                self.bottom_hints = [hint_refresh, hint_copy, hint_quit]
                members = self.iam_service.get_role_members(self.table, self.selected_value)
                self.content = output.get_members_output(members)

    def _update_panel(self, live: Live) -> None:
        window_layout = Layout(name="window")
        body_layout = Layout(name="body")
        hints_layout = output.hints_layout(self.hints, self.bottom_hints)
        self.content_panel = output.content_panel(self.content)
        content_layout = Layout(self.content_panel, name="content")
        list_layout = output.list_layout(self.values, self.selected_value)
        header_layout = output.header_layout(self.config)
        body_layout.split_row(hints_layout, content_layout, list_layout)
        window_layout.split_column(header_layout, body_layout)
        self.window_panel = output.window_panel(window_layout, self.now)
        self.layout = Layout(self.window_panel)
        live.update(self.layout, refresh=True)

    def _update_table(self, table: Optional[bigquery.Table]):
        """
        Update table identifiers to the state, save table to the history
        """
        if table:
            self.table = table
            self.project_id = table.project
            self.dataset_id = table.dataset_id
            self.history_service.save_table(table)

    def _loop(self, live: Live):
        """
        Loop listening on specific keypress, updating live CLI
        """
        self._update_content(live)
        self._update_panel(live)
        char = readchar.readkey()
        match char:
            case "t" | key.BACKSPACE | key.ESC:
                logger.trace("Table view")
                self.view = View.table
                self.values = []
                self.selected_value = None

            case key.UP if self.values and self.selected_value:
                idx = max([0, self.values.index(self.selected_value) - 1])
                self.selected_value = self.values[idx]

            case key.DOWN if self.values and self.selected_value:
                idx = min([len(self.values) - 1, self.values.index(self.selected_value) + 1])
                self.selected_value = self.values[idx]

            # List projects / Open new table
            case "1" | "o":
                logger.trace(f"Pressed 'o' ({hint_open.name})")
                self.view = View.table
                self.table = self.table_service.get_table(live=live)

            # List datasets
            case "2":
                logger.trace("Pressed '2' (List datasets)")
                table = self.table_service.get_table(self.project_id, live=live)
                if table:
                    self.view = View.table
                    self.table = table

            # List tables
            case "3":
                logger.trace("Pressed '3' (List tables)")
                table = self.table_service.get_table(self.project_id, self.dataset_id, live=live)
                if table:
                    self.view = View.table
                    self.table = table

            # List history
            case "h":
                logger.trace(f"Pressed 'h' ({hint_history.name})")
                table = self.history_service.pick_table(live)
                if table:
                    self.view = View.table
                    self.table = table

            # Refresh content
            case "r" if self.table:
                logger.trace(f"Pressed 'r' ({hint_refresh.name})")
                flash_panel(live, self.layout, self.window_panel)
                self.now = datetime.utcnow()
                self.table = self.table_service.get_fresh_table(self.table)

            # Show schema view
            case "s" if self.table:
                logger.trace(f"Pressed 's' ({hint_schema.name})")
                self.values = []
                self.view = View.schema

            # Open in browser
            case "b" if self.table:
                logger.trace(f"Pressed 'b' ({hint_open_browser.name})")
                url = f"https://console.cloud.google.com/bigquery?p={self.table.project}&d={self.table.dataset_id}&t={self.table.table_id}&page=table"
                webbrowser.open(url)

            # Copy
            case "c" if self.table:
                logger.trace(f"Pressed 'c' ({hint_copy.name})")
                copy_content = None
                match self.view:
                    case View.metadata | View.snippets | View.iam:
                        copy_content = self.content
                    case View.table:
                        copy_content = table_utils.get_properties(self.table)
                if copy_content:
                    flash_content(live, self.layout, self.content_panel)
                    pyperclip.copy(copy_content)

            # Show metadata
            case "m" if self.table:
                logger.trace(f"Pressed 'm' ({hint_metadata.name})")
                self.view = View.metadata
                self.values = [Metadata.table_id.name, Metadata.link.name, Metadata.schema.name]
                self.selected_value = Metadata.table_id.name

            # Show iam
            case "i" if self.table:
                logger.trace(f"Pressed 'i' ({hint_iam.name})")
                iam_result = self.iam_service.fetch_all_roles_members(self.table, self.config.iam_roles)
                if iam_result:
                    self.view = View.iam
                    self.values = self.config.iam_roles
                    self.selected_value = self.config.iam_roles[0]
                else:
                    self.view = View.error
                    self.content = output.get_missing_assets_permission_output(self.config, self.table)

            # Show snippets view
            case "p" if self.table:
                logger.trace(f"Pressed 'p' ({hint_snippets.name})")
                self.view = View.snippets
                snippets = self.snippet_service.list_snippets()
                self.values = snippets
                self.selected_value = snippets[0] if snippets else None

            # Quit
            case "q":
                logger.trace(f"Pressed 'q' ({hint_quit.name})")
                live.stop()
                click.get_current_context().exit()

        # Infinite loop, until quitted (q)
        self._loop(live)



from typing import List

from loguru import logger
from rich.console import Console

from bq_meta.client import Client
from bq_meta.config import Config
from bq_meta.util.rich_utils import progress
from google.cloud.bigquery import Table

from google.cloud.asset_v1 import AnalyzeIamPolicyRequest, IamPolicyAnalysisQuery


class IamService:
    def __init__(self, console: Console, config: Config, client: Client) -> None:
        self.console = console
        self.config = config
        self.client = client
        self.table = None
        self.role_members = {}

    def fetch_all_roles_members(self, table: Table, roles: List[str]) -> bool:
        logger.trace("Method call")
        try:
            self.table = table
            request = AnalyzeIamPolicyRequest()
            analysis_query = IamPolicyAnalysisQuery()
            analysis_query.scope = f"projects/{table.project}"
            analysis_query.access_selector.roles = roles
            analysis_query.resource_selector.full_resource_name = (
                f"//bigquery.googleapis.com/projects/{table.project}/datasets/{table.dataset_id}/tables/{table.table_id}"
            )
            request.analysis_query = analysis_query
            logger.debug(request)
            response = self.client.asset_client.analyze_iam_policy(request=request)
            self.role_members = {}
            for result in response.main_analysis.analysis_results:
                self.role_members[result.iam_binding.role] = result.iam_binding.members
            logger.debug(f"Fetched all roles members: {self.role_members}")
            return True
        except Exception as e:
            logger.warning(e)
            return False
            

    def get_role_members(self, table: Table, role: str) -> List[str]:
        logger.trace("Method call")
        members = []
        if self.table.full_table_id == table.full_table_id:
            members = self.role_members[role]
        return members
